﻿#include "pch.h"
#include "CircularTextRenderer.h"
#include "Common/DirectXHelper.h"
#include "FlattenedGeometrySink.h"

using namespace CircularText;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;

// Initializes D2D resources used for text rendering.
CircularTextRenderer::CircularTextRenderer(const std::shared_ptr<DeviceResources>& deviceResources) : 
	m_deviceResources(deviceResources)
{
    // Load font file
    ComPtr<IDWriteFactory> factory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Miramo.ttf";
    ComPtr<IDWriteFontFile> fontFile;
    DX::ThrowIfFailed(
        factory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;
    DX::ThrowIfFailed(
        factory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
        1,
        fontFile.GetAddressOf(),
        0,
        DWRITE_FONT_SIMULATIONS_NONE,
        &fontFace)
        );

    // Convert string to glyph indices
    std::wstring str = L"ext in an Infinite Circle of T";
    uint32 glyphCount = str.length();
    std::vector<uint32> str32(glyphCount);

    for (uint32 i = 0; i < glyphCount; i++)
        str32[i] = str[i];

    std::vector<uint16> glyphIndices = std::vector<uint16>(glyphCount);

    DX::ThrowIfFailed(
        fontFace->GetGlyphIndices(str32.data(), glyphCount, glyphIndices.data())
        );

    // Get glyph outline in path geometry
    ComPtr<ID2D1PathGeometry> pathGeometry;
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&pathGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;
    DX::ThrowIfFailed(
        pathGeometry->Open(&geometrySink)
        );

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(100,
                                     glyphIndices.data(),
                                     nullptr,
                                     nullptr,
                                     glyphIndices.size(),
                                     false,
                                     false,
                                     geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Save path geometry bounds
    DX::ThrowIfFailed(
        pathGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

    // Obtain flattened path geometry in FlattenedGeometrySink
    ComPtr<FlattenedGeometrySink> flattenedGeometrySink = new FlattenedGeometrySink(true);

    DX::ThrowIfFailed(
        pathGeometry->Simplify(D2D1_GEOMETRY_SIMPLIFICATION_OPTION_LINES,
        IdentityMatrix(),
        flattenedGeometrySink.Get())
        );

    DX::ThrowIfFailed(
        flattenedGeometrySink->Close()
        );

    // Get source polygons 
    m_srcPolygons = flattenedGeometrySink->GetPolygons();

    // Initialize size of destination polygon collection
    size_t count = m_srcPolygons.size();
    m_dstPolygons = std::vector<Polygon>(count);

    for (size_t polygonIndex = 0; polygonIndex < m_srcPolygons.size(); polygonIndex++)
    {
        // Initialize size of all Points collections
        count = m_srcPolygons.at(polygonIndex).Points.size();
        m_dstPolygons.at(polygonIndex) = Polygon(count);
    }

    // Create initial path geometry so Render is happy
    DX::ThrowIfFailed(
        Polygon::CreateGeometry(m_deviceResources->GetD2DFactory(),
                                m_srcPolygons,
                                &m_pathGeometry)
        );

	// Create other device independent resources
	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

	CreateDeviceDependentResources();
}

void CircularTextRenderer::CreateDeviceDependentResources()
{
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );

    DX::ThrowIfFailed(
        m_deviceResources->GetD2DDeviceContext()->CreateSolidColorBrush(ColorF(ColorF::Blue), &m_blueBrush)
        );
}
void CircularTextRenderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
    m_blueBrush.Reset();
}

void CircularTextRenderer::CreateWindowSizeDependentResources()
{
    // Get window size and geometry size
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();
    float geometryWidth = m_geometryBounds.right - m_geometryBounds.left;
    float geometryHeight = m_geometryBounds.bottom - m_geometryBounds.top;

    // Calculate a few factors for converting 2D to 3D
    float radius = logicalSize.Width / 2 - 50;
    float circumference = 2 * 3.14159f * radius;
    float scale = circumference / geometryWidth;
    float height = scale * geometryHeight;

    for (size_t polygonIndex = 0; polygonIndex < m_srcPolygons.size(); polygonIndex++)
    {
        const Polygon& srcPolygon = m_srcPolygons.at(polygonIndex);
        Polygon& dstPolygon = m_dstPolygons.at(polygonIndex);

        for (size_t pointIndex = 0; pointIndex < srcPolygon.Points.size(); pointIndex++)
        {
            const D2D1_POINT_2F pt = srcPolygon.Points.at(pointIndex);
            float radians = 2 * 3.14159f * (pt.x - m_geometryBounds.left) / geometryWidth;
            float x = radius * sin(radians);
            float z = radius * cos(radians);
            float y = height * ((pt.y - m_geometryBounds.top) / geometryHeight - 0.5f);
            dstPolygon.Points.at(pointIndex) = Point2F(x, y);
        }
    }

    // Create path geometry from Polygon collection
    DX::ThrowIfFailed(
        Polygon::CreateGeometry(m_deviceResources->GetD2DFactory(),
                                m_dstPolygons,
                                &m_pathGeometry)
        );
}

void CircularTextRenderer::Update(DX::StepTimer const& timer)
{
}

// Renders a frame to the screen.
void CircularTextRenderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::LightPink));

    // Set transform to center of screen
    Matrix3x2F center = Matrix3x2F::Translation(logicalSize.Width / 2,
                                                logicalSize.Height / 2);

    context->SetTransform(center * m_deviceResources->GetOrientationTransform2D());

    // Display path geometry
    context->FillGeometry(m_pathGeometry.Get(), m_blueBrush.Get());
    context->DrawGeometry(m_pathGeometry.Get(), m_blackBrush.Get(), 3);

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

