﻿#include "pch.h"
#include "OccludedCircularTextRenderer.h"
#include "Common/DirectXHelper.h"
#include "FlattenedGeometrySink.h"

using namespace OccludedCircularText;

using namespace D2D1;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::ApplicationModel;
using namespace Windows::Foundation;

OccludedCircularTextRenderer::OccludedCircularTextRenderer(const std::shared_ptr<DeviceResources>& deviceResources) : 
	m_text(L""),
    m_tiltAngle(10),
    m_margin(50),
	m_deviceResources(deviceResources)
{
    // Load font file
    ComPtr<IDWriteFactory> factory = m_deviceResources->GetDWriteFactory();
    String^ filePath = Package::Current->InstalledLocation->Path + "\\Fonts\\Miramo.ttf";
    ComPtr<IDWriteFontFile> fontFile;
    DX::ThrowIfFailed(
        factory->CreateFontFileReference(filePath->Data(), nullptr, &fontFile)
        );

    // Create font face
    ComPtr<IDWriteFontFace> fontFace;
    DX::ThrowIfFailed(
        factory->CreateFontFace(DWRITE_FONT_FACE_TYPE_TRUETYPE,
                                1,
                                fontFile.GetAddressOf(),
                                0,
                                DWRITE_FONT_SIMULATIONS_NONE,
                                &fontFace)
        );

    // Convert string to glyph indices
    std::wstring str = L"ext in an Infinite Circle of T";
    uint32 glyphCount = str.length();
    std::vector<uint32> str32(glyphCount);

    for (uint32 i = 0; i < glyphCount; i++)
        str32[i] = str[i];

    std::vector<uint16> glyphIndices = std::vector<uint16>(glyphCount);

    DX::ThrowIfFailed(
        fontFace->GetGlyphIndices(str32.data(), glyphCount, glyphIndices.data())
        );

    // Get glyph outline in path geometry
    ComPtr<ID2D1PathGeometry> pathGeometry;
    DX::ThrowIfFailed(
        m_deviceResources->GetD2DFactory()->CreatePathGeometry(&m_srcPathGeometry)
        );

    ComPtr<ID2D1GeometrySink> geometrySink;
    DX::ThrowIfFailed(
        m_srcPathGeometry->Open(&geometrySink)
        );

    DX::ThrowIfFailed(
        fontFace->GetGlyphRunOutline(100,
                                     glyphIndices.data(),
                                     nullptr,
                                     nullptr,
                                     glyphIndices.size(),
                                     false,
                                     false,
                                     geometrySink.Get())
        );

    DX::ThrowIfFailed(
        geometrySink->Close()
        );

    // Save path geometry bounds
    DX::ThrowIfFailed(
        m_srcPathGeometry->GetBounds(IdentityMatrix(), &m_geometryBounds)
        );

    // Create other device independent resources
	DX::ThrowIfFailed(
		m_deviceResources->GetDWriteFactory()->CreateTextFormat(
			L"Segoe UI",
			nullptr,
			DWRITE_FONT_WEIGHT_LIGHT,
			DWRITE_FONT_STYLE_NORMAL,
			DWRITE_FONT_STRETCH_NORMAL,
			32.0f,
			L"en-US",
			&m_textFormat
			)
		);

	DX::ThrowIfFailed(
		m_textFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_FAR)
		);

    DX::ThrowIfFailed(
        m_textFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_TRAILING)
        );

	DX::ThrowIfFailed(
		m_deviceResources->GetD2DFactory()->CreateDrawingStateBlock(&m_stateBlock)
		);

	CreateDeviceDependentResources();
}

void OccludedCircularTextRenderer::CreateDeviceDependentResources()
{
    ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush)
        );

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(ColorF::Blue), &m_blueBrush)
        );

    DX::ThrowIfFailed(
        context->CreateSolidColorBrush(ColorF(0.5f, 0.5f, 0.5f), &m_grayBrush)
        );
}

void OccludedCircularTextRenderer::ReleaseDeviceDependentResources()
{
    m_blackBrush.Reset();
    m_blueBrush.Reset();
    m_grayBrush.Reset();
}

void OccludedCircularTextRenderer::StartTracking()
{
    m_tracking = true;
}

void OccludedCircularTextRenderer::TrackingUpdate(Point ptDelta)
{
    if (m_tracking)
    {
        m_tiltAngle -= ptDelta.Y / 2.5f;
        m_tiltAngle = max(-45, min(45, m_tiltAngle));
    }
}

void OccludedCircularTextRenderer::StopTracking()
{
    m_tracking = false;
}

// Updates the text to be displayed.
void OccludedCircularTextRenderer::Update(DX::StepTimer const& timer)
{
    // Get geometry size
    float geometryWidth = m_geometryBounds.right - m_geometryBounds.left;
    float geometryHeight = m_geometryBounds.bottom - m_geometryBounds.top;

    // Calculate rotation angle and front/back bounds
    float rotateAngle = -360 * float(fmod(timer.GetTotalSeconds(), 10)) / 10;
    float x1 = geometryWidth * fmodf(5.0f / 4 + rotateAngle / 360, 1);
    float x2 = geometryWidth * fmodf(7.0f / 4 + rotateAngle / 360, 1);

    // Determine front and back mask rectangle
    D2D1_RECT_F maskRect;
    D2D1_COMBINE_MODE frontCombineMode;
    D2D1_COMBINE_MODE backCombineMode;

    if (x1 < x2)
    {
        maskRect = RectF(x1, m_geometryBounds.top, x2, m_geometryBounds.bottom);
        frontCombineMode = D2D1_COMBINE_MODE_EXCLUDE;
        backCombineMode = D2D1_COMBINE_MODE_INTERSECT;
    }
    else
    {
        maskRect = RectF(x2, m_geometryBounds.top, x1, m_geometryBounds.bottom);
        frontCombineMode = D2D1_COMBINE_MODE_INTERSECT;
        backCombineMode = D2D1_COMBINE_MODE_EXCLUDE;
    }

    ComPtr<ID2D1RectangleGeometry> maskGeometry;
    m_deviceResources->GetD2DFactory()->CreateRectangleGeometry(maskRect, &maskGeometry);

    // Create front geometry
    DX::ThrowIfFailed(
        CreateGeometry(maskGeometry.Get(), frontCombineMode, rotateAngle, &m_frontPathGeometry)
        );

    // Create back geometry
    DX::ThrowIfFailed(
        CreateGeometry(maskGeometry.Get(), backCombineMode, rotateAngle, &m_backPathGeometry)
        );

    // Update display text.
    uint32 fps = timer.GetFramesPerSecond();

    m_text = (fps > 0) ? std::to_wstring(fps) + L" FPS" : L" - FPS";
}

HRESULT OccludedCircularTextRenderer::CreateGeometry(ID2D1Geometry * maskGeometry, 
                                                     D2D1_COMBINE_MODE combineMode, 
                                                     float rotationAngle,
                                                     ID2D1PathGeometry ** pathGeometry)
{
    HRESULT hr;
    ComPtr<ID2D1PathGeometry> maskedPathGeometry;
    
    if (FAILED(
        hr = m_deviceResources->GetD2DFactory()->CreatePathGeometry(&maskedPathGeometry)
        ))
        return hr;

    ComPtr<ID2D1GeometrySink> geometrySink;

    if (FAILED(
        hr = maskedPathGeometry->Open(&geometrySink)
        ))
        return hr;

    m_srcPathGeometry->CombineWithGeometry(maskGeometry, combineMode, IdentityMatrix(), geometrySink.Get());

    if (FAILED(
        hr = geometrySink->Close()
        ))
        return hr;

    // Obtain flattened path geometry in FlattenedGeometrySink
    ComPtr<FlattenedGeometrySink> flattenedGeometrySink = new FlattenedGeometrySink(true);

    if (FAILED(
        hr = maskedPathGeometry->Simplify(D2D1_GEOMETRY_SIMPLIFICATION_OPTION_LINES,
        IdentityMatrix(),
        flattenedGeometrySink.Get())
        ))
        return hr;

    if (FAILED(
        hr = flattenedGeometrySink->Close()
        ))
        return hr;

    // Get source polygons 
    std::vector<Polygon> srcPolygons = flattenedGeometrySink->GetPolygons();

    // Initialize size of destination polygon collection
    size_t count = srcPolygons.size();
    std::vector<Polygon> dstPolygons(count);

    for (size_t polygonIndex = 0; polygonIndex < srcPolygons.size(); polygonIndex++)
    {
        // Initialize size of all Points collections
        count = srcPolygons.at(polygonIndex).Points.size();
        dstPolygons.at(polygonIndex) = Polygon(count);
    }

    // Get window size and geometry size
    Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();
    float geometryWidth = m_geometryBounds.right - m_geometryBounds.left;
    float geometryHeight = m_geometryBounds.bottom - m_geometryBounds.top;

    // Calculate a few factors for converting 2D to 3D
    float radius = logicalSize.Width / 2 - m_margin;
    float circumference = 2 * 3.14159f * radius;
    float scale = circumference / geometryWidth;
    float height = scale * geometryHeight;

    // Calculate rotation matrix
    Matrix3x2F rotateMatrix = Matrix3x2F::Rotation(rotationAngle);

    // Calculate tilt matrix
    Matrix3x2F tiltMatrix = Matrix3x2F::Rotation(m_tiltAngle);

    for (size_t polygonIndex = 0; polygonIndex < srcPolygons.size(); polygonIndex++)
    {
        const Polygon& srcPolygon = srcPolygons.at(polygonIndex);
        Polygon& dstPolygon = dstPolygons.at(polygonIndex);

        for (size_t pointIndex = 0; pointIndex < srcPolygon.Points.size(); pointIndex++)
        {
            const D2D1_POINT_2F pt = srcPolygon.Points.at(pointIndex);
            float radians = 2 * 3.14159f * (pt.x - m_geometryBounds.left) / geometryWidth;
            float x = radius * sin(radians);
            float z = radius * cos(radians);
            float y = height * ((pt.y - m_geometryBounds.top) / geometryHeight - 0.5f);

            // Apply rotation to X and Z
            D2D1_POINT_2F rotatedPoint = rotateMatrix.TransformPoint(Point2F(x, z));
            x = rotatedPoint.x;
            z = rotatedPoint.y;

            // Apply tilt to Y and Z
            D2D1_POINT_2F tiltedPoint = tiltMatrix.TransformPoint(Point2F(y, z));
            y = tiltedPoint.x;
            z = tiltedPoint.y;

            dstPolygon.Points.at(pointIndex) = Point2F(x, y);
        }
    }

    // Create path geometry from Polygon collection
    return Polygon::CreateGeometry(m_deviceResources->GetD2DFactory(),
        dstPolygons,
        pathGeometry);
}

// Renders a frame to the screen.
void OccludedCircularTextRenderer::Render()
{
	ID2D1DeviceContext* context = m_deviceResources->GetD2DDeviceContext();
	Windows::Foundation::Size logicalSize = m_deviceResources->GetLogicalSize();

	context->SaveDrawingState(m_stateBlock.Get());
	context->BeginDraw();
    context->Clear(ColorF(ColorF::LightPink));

    // Set transform to center of screen
    Matrix3x2F center = Matrix3x2F::Translation(logicalSize.Width / 2,
        logicalSize.Height / 2);

    context->SetTransform(center * m_deviceResources->GetOrientationTransform2D());

    // Display path geometries
    context->FillGeometry(m_backPathGeometry.Get(), m_grayBrush.Get());
    context->FillGeometry(m_frontPathGeometry.Get(), m_blueBrush.Get());
    context->DrawGeometry(m_frontPathGeometry.Get(), m_blackBrush.Get(), 3);

    // Display FPS value
    context->SetTransform(m_deviceResources->GetOrientationTransform2D());
    D2D1_RECT_F layoutRect = RectF(0, 0, logicalSize.Width, logicalSize.Height);
    context->DrawText(m_text.c_str(), m_text.length(), m_textFormat.Get(), &layoutRect, m_blackBrush.Get());

	// Ignore D2DERR_RECREATE_TARGET here. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = context->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}

	context->RestoreDrawingState(m_stateBlock.Get());
}

