#include "pch.h"
#include "FlattenedGeometrySink.h"

using namespace D2D1;
using namespace DirectX;

// IUnknown methods
ULONG STDMETHODCALLTYPE FlattenedGeometrySink::AddRef()
{
    m_refCount++;
    return m_refCount;
}

ULONG STDMETHODCALLTYPE FlattenedGeometrySink::Release()
{
    m_refCount--;
    LONG newCount = m_refCount;

    if (m_refCount == 0)
        delete this;

    return newCount;
}

HRESULT STDMETHODCALLTYPE FlattenedGeometrySink::QueryInterface(_In_ REFIID riid, _Outptr_ void** ppOutput)
{
    *ppOutput = nullptr;
    HRESULT hr = S_OK;

    if (riid == __uuidof(ID2D1SimplifiedGeometrySink))
    {
        *ppOutput = static_cast<ID2D1SimplifiedGeometrySink*>(this);
        AddRef();
    }
    else if (riid == __uuidof(IUnknown))
    {
        *ppOutput = this;
        AddRef();
    }
    else
    {
        hr = E_NOINTERFACE;
    }
    return hr;
}

// ID2D1SimplifiedGeometrySink methods
FlattenedGeometrySink::FlattenedGeometrySink(bool interpolate, float flatteningTolerance) :
    m_refCount(0),
    m_interpolate(interpolate),
    m_flatteningTolerance(flatteningTolerance),
    m_hresult(S_OK)
{
}

void FlattenedGeometrySink::SetFillMode(D2D1_FILL_MODE fillMode)
{
    // Ignore
}

void FlattenedGeometrySink::SetSegmentFlags(D2D1_PATH_SEGMENT vertexFlags)
{
    // Ignore
}

void FlattenedGeometrySink::BeginFigure(D2D1_POINT_2F startPoint,
                                        D2D1_FIGURE_BEGIN figureBegin)
{
    m_polygon.Points.clear();
    m_polygon.Points.push_back(startPoint);
    m_prevPoint = startPoint;
}

void FlattenedGeometrySink::AddLines(const D2D1_POINT_2F *points,
                                     UINT pointsCount)
{
    if (!m_interpolate)
    {
        m_polygon.Points.insert(m_polygon.Points.end(), pointsCount, *points);
        m_prevPoint = points[pointsCount - 1];
    }
    else
    {
        for (UINT i = 0; i < pointsCount; i++)
        {
            D2D1_POINT_2F point = points[i];
            Interpolate(m_prevPoint, point, m_polygon.Points);
            m_prevPoint = point;
        }
    }
}

void FlattenedGeometrySink::AddBeziers(const D2D1_BEZIER_SEGMENT *beziers,
                                       UINT beziersCount)
{
    // Beziers are not valid for this geometry sink
    m_hresult = E_ILLEGAL_METHOD_CALL;
}

void FlattenedGeometrySink::EndFigure(D2D1_FIGURE_END figureEnd)
{
    if (figureEnd == D2D1_FIGURE_END_CLOSED)
    {
        m_polygon.Points.push_back(m_polygon.Points[0]);
    }

    m_polygon.Points.shrink_to_fit();
    m_polygons.push_back(std::move(m_polygon));
}

HRESULT FlattenedGeometrySink::Close()
{
    return m_hresult;
}

// Methods for this implementation
std::vector<Polygon> FlattenedGeometrySink::GetPolygons()
{
    m_polygons.shrink_to_fit();
    return std::move(m_polygons);
}

void FlattenedGeometrySink::Interpolate(D2D1_POINT_2F point0,
                                        D2D1_POINT_2F point1,
                                        std::vector<D2D1_POINT_2F>& points)
{
    XMVECTOR pt0 = XMVectorSet(point0.x, point0.y, 0, 0);
    XMVECTOR pt1 = XMVectorSet(point1.x, point1.y, 0, 0);
    XMVECTOR vector = pt1 - pt0;

    float length = XMVectorGetX(XMVector2Length(vector));

    if (length < m_flatteningTolerance)
    {
        points.push_back(point1);
    }
    else
    {
        int count = (int) (length / m_flatteningTolerance) + 1;
        XMVECTOR increment = vector / float(count);

        for (int i = 0; i < count; i++)
        {
            pt0 += increment;
            points.push_back(Point2F(XMVectorGetX(pt0), XMVectorGetY(pt0)));
        }
    }
}
